"""
Benchmarking tools.
"""

# Future Library
from __future__ import annotations

# Standard Library
import logging
import time

from inspect import signature
from itertools import product
from typing import (
    Any,
    Callable,
    Iterable,
    Literal,
    Optional,
    ParamSpec,
    TypeVar,
    Union,
    cast,
    get_args,
)

# 3rd Party Library
import numpy as np
import pandas as pd
import plotnine as pn

from icontract import require
from numpy.typing import ArrayLike
from scipy.stats import ttest_ind
from tqdm.auto import (  # This works better in Jupyter notebooks than 'from tqdm import tqdm'
    tqdm,
)

# 1st Party Library
from rwskit.collections_ import is_generator, is_iterable

log = logging.getLogger(__name__)


P = ParamSpec("P")
"""A generic type for the parameters of a callable."""

R = TypeVar("R")
"""A generic type for the return value of a callable."""


__all__ = [
    "TimeUnit",
    "AggregationFunctionName",
    "AggregationFunction",
    "BenchmarkSortValue",
    "BenchmarkResult",
    "BenchmarkRunner",
    "change_time_unit",
    "get_time_unit_abbreviation",
    "validate_call_signature",
]

T = TypeVar("T", bool, int, float, str)
I = TypeVar("I")

TimeUnit = Literal[
    "seconds",
    "s",
    "milliseconds",
    "ms",
    "microseconds",
    "us",
    "µs",
    "nanoseconds",
    "ns",
]
"""
The supported units of time.
"""

AggregationFunctionName = Literal["min", "max", "mean", "median", "sum"]
"""
The names of the supported aggregation functions.
"""

AggregationFunction = Callable[[ArrayLike], float]
"""
An aggregation function is a callable that takes an array-like object and
returns a single float.
"""

BenchmarkSortValue = Literal["min", "mean", "function"]
"""
The supported values for sorting the ``BenchmarkResults`` when represented
as a string.
"""

PlotTheme = Literal[
    "theme_538",
    "theme_bw",
    "theme_classic",
    "theme_dark",
    "theme_gray",
    "theme_grey",
    "theme_light",
    "theme_linedraw",
    "theme_matplotlib",
    "theme_minimal",
    "theme_seaborn",
    "theme_tufte",
    "theme_void",
    "theme_xkcd",
]

_time_conversion_factors = {
    "seconds": 1,
    "milliseconds": 1e-3,
    "microseconds": 1e-6,
    "nanoseconds": 1e-9,
}

_time_unit_abbreviation_to_name: dict[TimeUnit, TimeUnit] = {
    "s": "seconds",
    "ms": "milliseconds",
    "us": "microseconds",
    "µs": "microseconds",
    "ns": "nanoseconds",
    "seconds": "seconds",
    "milliseconds": "milliseconds",
    "microseconds": "microseconds",
    "nanoseconds": "nanoseconds",
}

_time_unit_name_to_abbreviation = {
    "seconds": "s",
    "milliseconds": "ms",
    "microseconds": "µs",
    "nanoseconds": "ns",
}

_valid_themes = {
    pn.theme_538,
    pn.theme_bw,
    pn.theme_classic,
    pn.theme_dark,
    pn.theme_gray,
    pn.theme_grey,
    pn.theme_light,
    pn.theme_linedraw,
    pn.theme_matplotlib,
    pn.theme_minimal,
    pn.theme_seaborn,
    pn.theme_tufte,
    pn.theme_void,
    pn.theme_xkcd,
}

_theme_name_to_theme = {theme.__name__: theme for theme in _valid_themes}


@require(
    lambda name: name in _time_unit_abbreviation_to_name.keys(),
    f"Unsupported time unit. Must be one of: {_time_unit_abbreviation_to_name.keys()}",
)
def get_time_unit_abbreviation(name: TimeUnit) -> TimeUnit:
    """
    Get the time unit abbreviation from the given string.

    Parameters
    ----------
    name : str
        The name or abbreviation of a supported time unit.

    Returns
    -------
    str
        The abbreviation of the time unit specified by ``name``.

    Raises
    ------
    icontract.errors.ViolationError
        If the ``name`` is not a supported :data:`TimeUnit`.

    """
    return _time_unit_abbreviation_to_name.get(name, name)


@require(lambda from_unit: from_unit in get_args(TimeUnit), "Invalid time unit")
@require(lambda to_unit: to_unit in get_args(TimeUnit), "Invalid time unit")
def change_time_unit(
    value: int | float, from_unit: TimeUnit, to_unit: TimeUnit
) -> float:
    """
    Change the unit of time of a given ``value`` currently in the ``from_unit``
    unit to a value in the ``to_unit`` unit.

    Parameters
    ----------
    value : int or float
        The current time value.
    from_unit : TimeUnit
        The unit of the current value.
    to_unit : TimeUnit
        The unit to change the value into.

    Returns
    -------
    float
        Return the equivalent value in the new time unit ``to_unit``
    """
    seconds = value * _time_conversion_factors[from_unit]

    return seconds / _time_conversion_factors[to_unit]


def validate_call_signature(
    fn1: Callable[..., Any], fn2: Callable[..., Any], strict: bool = False
) -> bool:
    """
    Check that the two functions take the same parameters.

    Parameters
    ----------
    fn1 : Callable[..., Any]
        The first function to compare.
    fn2 : Callable[..., Any]
        The second function to compare.
    strict : bool, default = False
        If ``True`` the signatures must match exactly, including whether
        defaults are present and their values. Otherwise, they are considered
        equal if the number and types of all parameters are the same.

    Returns
    -------
    bool
        True if the functions take the same number and type of parameters.
    """
    params1 = signature(fn1).parameters.values()
    params2 = signature(fn2).parameters.values()

    if len(params1) != len(params2):
        return False

    # Strict requires all 'Parameter' objects between the two functions to be
    # equal. They are equal if the name, kind, default value, and annotation
    # are the same.
    if strict:
        return all(p1 == p2 for p1, p2 in zip(params1, params2))

    # Non-strict only requires the name and annotation to be the same.
    return all(
        p1.name == p2.name and p1.annotation == p2.annotation
        for p1, p2 in zip(params1, params2)
    )


# region BenchmarkRunner Input Validation
def _are_setup_fn_and_functions_compatible(
    benchmark_space: dict[str, list[T]],
    setup_fn: Optional[Callable],
    functions: Iterable[Callable],
) -> bool:
    if setup_fn is None:
        return True

    functions = (
        list(functions.values()) if isinstance(functions, dict) else list(functions)
    )
    kwargs = {k: v[0] for k, v in benchmark_space.items()}
    setup_fn_output = setup_fn(**kwargs)

    if not isinstance(setup_fn_output, dict):
        return False

    parameter_names = [p.name for p in signature(functions[0]).parameters.values()]

    if len(parameter_names) != len(setup_fn_output):
        return False

    return all(k in parameter_names for k in setup_fn_output.keys())


def _are_benchmark_space_and_functions_compatible(
    setup_fn: Optional[Callable],
    benchmark_space: dict[str, list[T]],
    functions: Iterable[Callable],
) -> bool:
    if setup_fn is not None:
        return True

    functions = (
        list(functions.values()) if isinstance(functions, dict) else list(functions)
    )

    if not bool(functions):
        # We shouldn't get here if the decorators are ordered properly
        return False

    parameter_names = [p.name for p in signature(functions[0]).parameters.values()]

    if len(parameter_names) != len(benchmark_space):
        return False

    return all(k in parameter_names for k in benchmark_space)


def _is_space_distinct_from_run_label(
    benchmark_space: dict[str, list[T]], run_label: str
) -> bool:
    return not any(k == run_label for k in benchmark_space.keys())


def _functions_have_same_call_signatures(functions: Iterable[Callable]) -> bool:
    functions = (
        list(functions.values()) if isinstance(functions, dict) else list(functions)
    )
    if len(functions) < 1:
        return False

    fn1 = functions[0]

    return all(validate_call_signature(fn1, fn2) for fn2 in functions[1:])


def _functions_have_proper_type(functions: Iterable[Callable]) -> bool:
    if isinstance(functions, dict):
        return all(isinstance(k, str) and callable(v) for k, v in functions.items())

    return all(callable(f) for f in functions)


def _are_benchmark_space_and_setup_fn_compatible(
    benchmark_space: dict[str, list[T]], setup_fn: Optional[Callable]
) -> bool:
    # There's nothing to check
    if setup_fn is None:
        return True

    parameter_names = [p.name for p in signature(setup_fn).parameters.values()]

    # They need the same number of arguments
    if len(parameter_names) != len(benchmark_space):
        return False

    return all(k in parameter_names for k in benchmark_space)


def _benchmark_space_has_valid_keys(benchmark_space: dict[str, list[T]]) -> bool:
    return not any(k in ("min", "max", "mean", "std") for k in benchmark_space)


def _benchmark_space_is_dict_of_lists(benchmark_space: dict[str, list[T]]) -> bool:
    is_dict_of_lists = all(
        isinstance(k, str) and isinstance(v, list) for k, v in benchmark_space.items()
    )
    is_non_empty = bool(benchmark_space) and any(
        bool(v) for v in benchmark_space.values()
    )
    has_valid_values = all(
        isinstance(v, (bool, int, float, str))
        for lov in benchmark_space.values()
        for v in lov
    )

    return is_dict_of_lists and is_non_empty and has_valid_values


def _is_valid_float_format(float_fmt: str) -> bool:
    try:
        format(123.45, float_fmt)
    except (ValueError, TypeError):
        return False
    else:
        return True


# endregion BenchmarkRunner Input Validation


class BenchmarkRunner:
    """
    A class for comparing the execution time of multiple functions
    """

    _aggregate_function_lookup = {
        "min": np.min,
        "max": np.max,
        "mean": np.mean,
        "median": np.median,
        "sum": np.sum,
    }

    # Decorators are evaluated from bottom to top. So more complex contracts,
    # especially ones that depend on previous checks being performed, must be
    # defined first even though it is more intuitive to have the contracts
    # provided in the same order as the __init__ arguments.
    @require(
        _are_setup_fn_and_functions_compatible,
        "'setup_fn' should return a dict suitable to use as **kwargs for the test functions.",
    )
    @require(
        _are_benchmark_space_and_functions_compatible,
        "The 'benchmark_space' keys must match the 'functions' params when 'setup_fn' is None.",
    )
    @require(
        _is_space_distinct_from_run_label,
        "The 'benchmark_space' can't have a key that is equal to the 'run_label'.",
    )
    @require(
        _functions_have_same_call_signatures,
        "All the functions must have the same call signature.",
    )
    @require(
        _functions_have_proper_type,
        "'functions' must be a dict[str, callable] or an iterable of callables.",
    )
    @require(lambda functions: bool(list(functions)), "'functions' must be non-empty.")
    @require(
        lambda functions: not is_generator(functions),
        "'functions' cannot be a generator.",
    )
    @require(lambda functions: is_iterable(functions), "'functions' must be iterable.")
    @require(
        _are_benchmark_space_and_setup_fn_compatible,
        "The 'benchmark_space' keys must match the 'setup_fn' params if 'setup_fn' is given.",
    )
    @require(
        lambda setup_fn: callable(setup_fn) if setup_fn is not None else True,
        "'setup_fn' must be callable.",
    )
    @require(
        _benchmark_space_has_valid_keys,
        "'benchmark_space' cannot contain the keys: [min, max, mean, std]",
    )
    @require(
        _benchmark_space_is_dict_of_lists,
        "'benchmark_space' must be a non-empty dict[str, list[bool, int, float, str]].",
    )
    @require(lambda n_runs: n_runs > 0, "There must be at least one run.")
    @require(lambda n_tests: n_tests > 0, "There must be at least one test")
    @require(lambda n_warm_ups: n_warm_ups >= 0, "There can't be negative warm ups.")
    @require(
        lambda time_unit: time_unit in get_args(TimeUnit),
        f"Invalid 'time_unit' must be one of: {get_args(TimeUnit)}",
    )
    @require(
        lambda test_agg_fn: test_agg_fn in get_args(AggregationFunctionName),
        f"Invalid 'test_agg_fn' function name. It must be one of: {get_args(AggregationFunctionName)}",
    )
    @require(_is_valid_float_format, "Invalid 'float_fmt'")
    @require(
        lambda sort_by: sort_by in get_args(BenchmarkSortValue),
        f"Invalid 'sort_by' value. It must be one of: {get_args(BenchmarkSortValue)}.",
    )
    def __init__(
        self,
        functions: Iterable[Callable] | dict[str, Callable],
        benchmark_space: dict[str, list[T]],
        setup_fn: Optional[Callable] = None,
        use_single_setup: bool = True,
        n_runs: int = 10,
        n_tests: int = 2,
        n_warm_ups: int = 1,
        time_unit: TimeUnit = cast(TimeUnit, "s"),
        test_agg_fn: AggregationFunctionName = "min",
        run_label: str = "run",
        show_progress: bool = False,
        verbose: bool = True,
        float_fmt: str = "0.4e",
        sort_by: BenchmarkSortValue = "min",
        test_significance: bool = True,
    ):
        """
        A class for profiling a set of functions based on one or mor criteria.

        The high level view of the benchmarking process is as follows.
        For every combination of parameters in the ``benchmark_space`` a
        sub-benchmark will be run.
        There are 2 nested execution loops for each sub-benchmark.
        The innermost loop runs each function on the current data ``n_tests``
        number of times and aggregates the results using the ``test_agg_fn``.
        The same data is always used for this loop no matter what. The outer
        loop will run this process ``n_runs`` times. If ``use_single_setup``
        is ``True`` then the setup function will only be called once and
        will be used for all the runs. If it is ``False`` the setup
        function will be called for every run. The execution times for all
        runs will be stored in a Pandas DataFrame that can be retrieved after
        calling the benchmark.

        .. note::
            ``min``, ``max``, ``mean``, ``std`` cannot be used as keys in
            the ``benchmark_space``.

        .. note::
            ``functions`` must be an iterable, but cannot be a generator.

        .. note::
            The first parameter in the ``benchmark_space`` is always used
            as the x-axis for :meth:`BenchmarkResult.plot`.

        Parameters
        ----------
        functions : list[BenchmarkFunction]
            A list of functions to benchmark or a dictionary that maps a label
            to a benchmark function.
        benchmark_space : dict[str, list[T]]
            The space of values to benchmark over. A benchmark will be
            executed for each combination of values obtained from the
            dictionary. The combinations are formed by taking the Cartesian
            product taking one value from each list in the dictionary.
            The names of the keys of this dictionary must either be the names
            of keyword arguments of the ``setup_fn``, or keyword arguments of
            the benchmark functions if no ``setup_fn`` is provided. Only
            ``bool``, ``int``, ``float``, and ``str`` values are supported.
        setup_fn : SetupFunction
            A function that initializes data to be passed to the benchmark
            ``functions``. If ``None``, the values from ``setup_args`` will
            be passed directly to each function in ``functions``.
        use_single_setup : bool, default = True
            For functions that are guaranteed to be deterministic no matter
            what the input is, this should be ``True``. However, if the
            function is non-deterministic or the performance might depend
            on how the data is initialized, this should be ``False``.
        n_runs : int
            The number of execution tests to run.
        n_tests : int
            The number times to run each function in a single test.
        n_warm_ups : int
            The number of tests to run before recording the timing data.
        test_agg_fn : {'min', 'max', 'mean', 'median', 'sum'}
            The function to use for aggregating individual test results within
            a run.
        run_label : str
            The column label in the resulting Pandas ``DataFrame`` that
            indicates the run number for the given execution times.
        show_progress : bool = False
            Show progress bars while running the benchmark.
        verbose : bool, default = True
            Print the full results and summary statistics to ``stdout``
            when complete.
        float_fmt : str
            The format used to print floating point values to a string.
        sort_by : str {min, mean, function}
            When ``verbose=True`` this will determine how the results are
            sorted (either by the min run time, max run time or by the
            function name).
        test_significance: bool, default = False
            If ``True``, test if the difference in run times are different
            between all pairs of models.

        Notes
        -----

        **Deterministic Function and Deterministic Data**

        If your algorithm is deterministic and is not influenced at all
        by the content of the data, only its size, then I would suggest the
        following parameters:

        * ``use_single_setup = True``: Use the same data for all the runs
          on the current setup parameters.
        * ``n_runs > 1``: Run it at least a few times per parameter set
          to make sure there weren't any anomalies biasing the results.
        * ``n_tests = 1``: You should not need to run multiple tests here.

        **Deterministic Function and Non-Deterministic Data**

        If your function is deterministic (the sequence of execution is always
        the same), but could be influenced by the content of the data I would
        suggest the following parameters:

        * ``setup_fn != None``: The setup function should return different
          data each run (of the same size)
        * ``use_single_setup = False``: Run the setup function to generate
          new data on each run.
        * ``n_tests > 1``: Run the function on the same data a few times
          in case there was an anomaly, which could bias the result.
        * ``n_runs > 1``: Run the function on multiple different data
          sets to estimate how much variability is expected due to the
          makeup of the data.
        * ``test_agg_fn = 'min'``: Since the function should execute
          the same way on the same data, the `min` should be the most
          informative.

        **Non-Deterministic Function**

        If the function itself is non-deterministic you probably want something
        similar to the deterministic case with non-deterministic data. In
        this case however, it is probably pointless to set ``n_tests > 1``
        and you should just increase ``n_runs`` to get better overall estimates.

        Examples
        --------

        .. code-block:: python

            >>> import time
            >>> sort_setup_fn = (
            ...    lambda array_size, dtype, unique_values:
            ...        np.random.randint(unique_values, size=array_size).astype(dtype)
            ... )

            >>> b = BenchmarkRunner(functions={"fn1": lambda a: time.sleep(0.01),
            ...                                "fn2": lambda a: time.sleep(0.02)},
            ...                     benchmark_space={"array_size": [100, 10000],
            ...                                      "dtype": ["U", "int"],
            ...                                      "unique_values": [10, 100, 1000]},
            ...                     setup_fn=sort_setup_fn
            ...                     time_unit="ms"
            ...                     float_fmt="0.3f")

            >>> b()
            function  array_size  unique_values    min   mean    std
            --------------------------------------------------------
                 fn1         100             10  1.040  1.053  0.008
                 fn2         100             10  5.057  5.058  0.001
            --------------------------------------------------------
                 fn1         100            100  1.053  1.056  0.002
                 fn2         100            100  5.057  5.058  0.001
            --------------------------------------------------------
                 fn1         100           1000  1.056  1.057  0.000
                 fn2         100           1000  5.058  5.058  0.000
            --------------------------------------------------------
                 fn1       10000             10  1.056  1.057  0.000
                 fn2       10000             10  5.058  5.059  0.000
            --------------------------------------------------------
                 fn1       10000            100  1.056  1.057  0.000
                 fn2       10000            100  5.058  5.064  0.009
            --------------------------------------------------------
                 fn1       10000           1000  1.056  1.057  0.001
                 fn2       10000           1000  5.063  5.066  0.002



        """
        self._check_for_configuration_problems(use_single_setup, setup_fn, n_runs)

        self.functions = self._normalize_functions(functions)
        self.benchmark_space = benchmark_space
        self.setup_fn = setup_fn or self._passthrough_setup_args
        self.use_single_setup = use_single_setup
        self.n_runs = n_runs
        self.n_tests = n_tests
        self.n_warm_ups = n_warm_ups
        self.test_agg_fn = self._aggregate_function_lookup[test_agg_fn]
        self.run_label = run_label
        self.show_progress = show_progress
        self.verbose = verbose
        self.time_unit: TimeUnit = cast(TimeUnit, get_time_unit_abbreviation(time_unit))
        self.float_fmt = float_fmt
        self.sort_by: BenchmarkSortValue = sort_by
        self.test_significance = test_significance

    def __call__(self) -> BenchmarkResult:
        """Runs the benchmark.

        Returns
        -------
        BenchmarkResult

        """
        return self.run()

    def run(self) -> BenchmarkResult:
        """Runs the benchmark.

        Returns
        -------
        BenchmarkResult
        """
        # Create the DataFrame to store the results
        results = self._initialize_data_frame()

        # Create a generator to iterate through all possible combinations of
        # the benchmark space.
        parameter_values = product(*self.benchmark_space.values())

        # Set up the parameter space loop with optional progress indicators
        parameter_progress = self._progress(
            parameter_values,
            total=np.prod([len(x) for x in self.benchmark_space.values()]),
            position=0,
            unit="parameters",
        )

        # Loop through all combinations
        for parameter_values in parameter_progress:
            # Create the setup_fn kwargs
            setup_kwargs = {
                k: v for k, v in zip(self.benchmark_space.keys(), parameter_values)
            }

            # Get the function arguments here if we only want to generate
            # them once for all runs of this parameter set (i.e., pure deterministic functions).
            function_args = (
                self.setup_fn(**setup_kwargs) if self.use_single_setup else None
            )

            # Set up the run loop with optional progress indicators
            run_progress = self._progress(
                range(1, self.n_runs + 1), position=1, leave=False, unit="runs"
            )

            for run_num in run_progress:
                self._do_run(results, run_num, setup_kwargs, function_args)

        significance_results = (
            self._run_ttest(results) if self.test_significance else None
        )

        result = BenchmarkResult(
            results,
            significance_results,
            self.benchmark_space,
            self.float_fmt,
            self.sort_by,
            self.run_label,
            self.time_unit,
        )

        if self.verbose:
            print(result)  # noqa

        return result

    @staticmethod
    def _check_for_configuration_problems(
        use_single_setup: bool, setup_fn: Optional[Callable], n_runs: int
    ):
        if not use_single_setup and not setup_fn:
            log.warning(
                "Setting 'single_setup' to 'False' implies the performance of the functions "
                "may be non-deterministic. However, no 'setup_fn' was defined so be sure that "
                "the only non-determinism is coming from the functions themselves."
            )
        if not use_single_setup and n_runs == 1:
            log.warning(
                "Setting 'single_setup' to 'False' implies the performance of the functions "
                "may be non-deterministic. However, 'n_runs' is set to 1, which will not."
            )

    @staticmethod
    def _normalize_functions(
        functions: Union[Iterable[Callable[P, R]], dict[str, Callable[P, R]]],
    ) -> dict[str, Callable[P, R]]:
        if isinstance(functions, dict):
            return functions  # type: ignore
        else:
            return {f.__name__: f for f in functions}

    @staticmethod
    def _passthrough_setup_args(**kwargs: T) -> dict[str, T]:
        return kwargs

    @staticmethod
    def _identity(x: Any) -> Any:
        return x

    def _set_default_run_agg_fn(self) -> AggregationFunction:
        if self.use_single_setup:
            return self._aggregate_function_lookup["min"]

        return self._aggregate_function_lookup["mean"]

    def _progress(self, values: Iterable[I], **tqdm_kwargs):
        return tqdm(values, disable=not self.show_progress, **tqdm_kwargs)

    def _initialize_data_frame(self) -> pd.DataFrame:
        index_map = {
            self.run_label: list(range(1, self.n_runs + 1))
        } | self.benchmark_space

        multi_index = pd.MultiIndex.from_product(
            list(index_map.values()), names=index_map.keys()
        )

        return pd.DataFrame(
            index=multi_index, columns=list(self.functions.keys()), dtype=np.float64
        )

    def _do_run(
        self,
        results: pd.DataFrame,
        run_num: int,
        setup_args: dict[str, T],
        function_args: Optional[dict[str, T]],
    ):
        # If we didn't set the function arguments already (i.e., use_single_setup=False)
        # we need to set them here.
        function_args = function_args or self.setup_fn(**setup_args)

        for function_name, function in self.functions.items():
            row_index = (run_num,) + tuple(setup_args.values())
            column_name = function_name
            results.loc[row_index, column_name] = self._do_test(function, function_args)

    def _do_test(self, function: Callable, function_args: dict[str, Any]) -> float:
        test_times = [
            self._time_function(function, function_args)
            for _ in range(self.n_warm_ups + self.n_tests)
        ]

        return self.test_agg_fn(test_times[self.n_warm_ups :])

    def _time_function(
        self, function: Callable, function_args: dict[str, Any]
    ) -> float:
        start = time.perf_counter_ns()
        function(**function_args)
        return change_time_unit(
            time.perf_counter_ns() - start, "nanoseconds", self.time_unit
        )

    def _run_ttest(self, df: pd.DataFrame) -> pd.DataFrame:
        benchmark_space = list(self.benchmark_space)
        functions = list(self.functions.keys())

        def _pairwise_ttest(group):
            # pylance can't locate 'pvalue' even though it exists.
            data = {
                f1: {f2: ttest_ind(group[f1], group[f2]).pvalue for f2 in functions}  # type: ignore
                for f1 in functions
            }
            tt_df = pd.DataFrame(data)
            tt_df.index.name = "function"
            return tt_df

        return df.groupby(benchmark_space).apply(_pairwise_ttest).reset_index()


class BenchmarkResult:
    """A class for managing the results output by a :class:`BenchmarkRunner`."""

    def __init__(
        self,
        results: pd.DataFrame,
        significance_results: Optional[pd.DataFrame],
        benchmark_space: dict[str, list[T]],
        float_fmt: str = "0.4e",
        sort_by: BenchmarkSortValue = "min",
        run_label: str = "run",
        time_unit: TimeUnit = "s",
    ):
        """A class for managing the results output by a :class:`BenchmarkRunner`.

        .. note::
            This class is not intended to be instantiated directly.

        Parameters
        ----------
        results : DataFrame
            The results DataFrame obtained by a :class:`BenchmarkRunner`.
        significance_results : DataFrame
            Pairwise t-test results.
        benchmark_space : dict[string, list[T]]
            The parameters used to benchmark the functions.
        float_fmt : str
            A valid format string to use for floating point numbers.
        sort_by : str {min, mean}
            The summary statistic to sort the results by when represented
            as a string.
        run_label : str
            The label used to indicate the run number.
        time_unit : TimeUnit
            The original time unit used to benchmark the results.
        """
        self._results = results
        self._significance_results = significance_results
        self._benchmark_space = benchmark_space
        self._float_fmt = float_fmt
        self._sort_by = sort_by
        self._run_label = run_label
        self._time_unit: TimeUnit = time_unit
        self._function_names = set(results.columns)

    def __repr__(self) -> str:
        """The full pandas :class:`pandas.DataFrame` containing all the runs as a string.

        Returns
        -------
        str
            The full benchmark results as a string.

        """
        return self._results.to_string()

    def __str__(self) -> str:
        """Returns a table of the summary statistics of the benchmark results as a string.

        Returns
        -------
        str
            The summary statistics of the benchmark results as a string.

        """
        summary = self.summary()
        groups = summary.groupby(list(self.benchmark_space))
        widths = self._get_max_column_widths(summary)
        header = self._stringify_header(summary.columns, widths)
        header_size = len(header)

        lines = ["=" * header_size]
        lines += [header]
        for i, (key, group) in enumerate(groups):
            separator = "=" if i == 0 else "-"
            lines.append(separator * header_size)
            lines.extend(self._stringify_group(group, widths))

        return "\n".join(lines)

    @property
    def benchmark_space(self) -> dict[str, list[Any]]:
        """Return the parameters used to benchmark the functions.

        Returns
        -------
        dict[str, list[T]]
            The parameters used to produce these results.

        """
        return self._benchmark_space.copy()

    @require(
        lambda as_time_unit: (
            as_time_unit in get_args(TimeUnit) if as_time_unit else True
        )
    )
    def results(
        self, wide: bool = True, as_time_unit: Optional[TimeUnit] = None
    ) -> pd.DataFrame:
        """Get a pandas ``DataFrame`` containing the results.

        Parameters
        ----------
        as_time_unit : TimeUnit, optional
            Return the results in this time unit instead of the one used
            during the benchmark.
        wide : bool, default = True
            If ``True`` return the results in the default wide format, which
            is easier to read. Otherwise, return the results in long format,
            which can be easier to use for plotting.

        Returns
        -------
        DataFrame
            The DataFrame containing the results, either in wide or long format.

        """
        results = self._results.copy(deep=True)

        if as_time_unit is not None:
            results = results.map(
                lambda t: change_time_unit(t, self._time_unit, as_time_unit)
            )  # noqa

        if wide:
            return results

        return pd.melt(
            results.reset_index(),
            id_vars=[self._run_label] + list(self.benchmark_space.keys()),
            var_name="function",
            value_name="time",
        )

    @property
    def significance_results(self) -> pd.DataFrame:
        """Return the results of the significance tests as a :class:`pandas.DataFrame`.

        Returns
        -------
        DataFrame
            The results of the significance tests as a :class:`pandas.DataFrame`.

        """
        if self._significance_results is None:
            raise ValueError("Significance results have not been computed yet.")

        return self._significance_results.copy(deep=True)

    @require(
        lambda as_time_unit: (
            as_time_unit in get_args(TimeUnit) if as_time_unit else True
        )
    )
    def summary(
        self, wide: bool = True, as_time_unit: Optional[TimeUnit] = None
    ) -> pd.DataFrame:
        """The summary statistics of the benchmark results.

        Parameters
        ----------
        wide : bool, default = True
            If ``True``, return the results in wide format, otherwise, return
            the results in long format.
        as_time_unit : TimeUnit, optional
            Return the results in this time unit instead of the one used during
            the benchmark.

        Returns
        -------
        pd.DataFrame
            The DataFrame with the summary statistics.

        """
        long_results = self.results(wide=False, as_time_unit=as_time_unit)
        group_by = ["function"] + list(self.benchmark_space)
        summary = long_results.groupby(group_by).agg({
            "time": ["min", "max", "mean", "std"]
        })
        summary.columns = summary.columns = [c2 or c1 for c1, c2 in summary.columns]
        summary.reset_index(inplace=True)

        summary = self._add_significance_to_summary(summary)

        if wide:
            return summary

        return summary.melt(
            id_vars=["function"] + list(self.benchmark_space),
            var_name="agg",
            value_name="value",
        )

    # noinspection PyTypeChecker
    @require(
        lambda self, functions: (
            all(f in self._function_names for f in functions) if functions else True
        ),
        "At least one of the functions provided is not in the benchmark results.",
    )
    def plot(
        self,
        x_label: Optional[str] = None,
        use_stat: Literal["min", "mean"] = "min",
        functions: Iterable[str] = (),
        show_points: bool = False,
        show_ribbon: bool = False,
        free_y: bool = False,
        theme_name: Optional[PlotTheme] = None,
        figure_size: Optional[tuple[int, int]] = None,
        as_time_unit: Optional[TimeUnit] = None,
    ) -> pn.ggplot:
        """Create and return a `ggplot <https://plotnine.org/reference/ggplot.html#plotnine.ggplot>`__
        object that visualizes the benchmark results.

        .. note::
            Use ``show` or ``save``` on the resulting object to render or save it.

        Parameters
        ----------
        x_label : str, optional
            Label the `x-axis` with this value, if given. Otherwise, the first
            key in the ``benchmark_space`` will be used.
        use_stat : str {min, mean}
            Which stat to plot.
        functions: Iterable[str], optional
            If defined, limit the plot to these functions.
        show_points : bool, default = False
            Show each individual run as a point on the graph.
        show_ribbon : bool, default = False
            If ``True``, include a ribbon that encompasses the minimum and
            maximum value over all the runs in a group.
        free_y : bool, default = False
            If ``True`` and the ``benchmark_space`` includes more than 1
            parameter, the y-axis is not constrained to be the same for each
            resulting plot.
        theme_name : PlotTheme, optional
            The name of a theme to style your plot with, otherwise, it will
            use the default theme.
        figure_size : tuple[int, int], optional,
            Override the size of the resulting figure. If not specified and
            there are more than one ``benchmark_space`` parameters a heuristic
            is used to try to ensure each subgraph will be legible.
        as_time_unit : TimeUnit, optional
            Display the results in this time unit instead of the one used
            during the benchmarking.

        Returns
        -------
        plotnine.ggplot
            The plot object.

        """
        results = self.results(wide=False, as_time_unit=as_time_unit)
        summary = self.summary(as_time_unit=as_time_unit)
        summary["lb"] = summary["mean"] - summary["std"]
        summary["ub"] = summary["mean"] + summary["std"]

        if functions is not None:
            results = results[results["function"].isin(functions)]
            summary = summary[summary["function"].isin(functions)]

        search_space_names = list(self._benchmark_space.keys())
        x = search_space_names[0]
        theme = _theme_name_to_theme.get(theme_name)

        plot = pn.ggplot(summary)

        if show_points:
            plot += pn.geom_point(
                data=results, mapping=pn.aes(x=x, y="time", color="function")
            )

        plot += pn.geom_line(mapping=pn.aes(x=x, y=use_stat, color="function"))

        if show_ribbon:
            plot += pn.geom_ribbon(
                mapping=pn.aes(
                    x=x,
                    ymin="lb",
                    ymax="ub",
                    fill="function",
                ),
                inherit_aes=False,
                alpha=0.60,
            )

            plot += pn.geom_ribbon(
                mapping=pn.aes(
                    x=x,
                    ymin="min",
                    ymax="max",
                    fill="function",
                ),
                inherit_aes=False,
                alpha=0.20,
            )

        if x_label is not None:
            plot += pn.xlab(x_label)

        plot += pn.ylab(f"Time [{_time_unit_name_to_abbreviation[self._time_unit]}]")

        scales = "free_y" if free_y else "fixed"
        if len(search_space_names) == 2:
            plot += pn.facet_grid(
                rows=search_space_names[1], scales=scales, labeller="label_both"
            )
        elif len(search_space_names) == 3 and not free_y:
            plot += pn.facet_grid(
                rows=search_space_names[1],
                cols=search_space_names[2],
                labeller="label_both",
                scales=scales,
            )
        elif len(search_space_names) >= 3:
            ncol = len(self.benchmark_space[search_space_names[1]])
            plot += pn.facet_wrap(
                search_space_names[1:],
                ncol=ncol,
                scales=scales,
                labeller="label_both",
                dir="v",
            )

        if theme is not None:
            plot += theme()

        plot += pn.theme(
            figure_size=self._estimate_figure_size(figure_size),
        )

        return plot

    def _estimate_figure_size(
        self, figure_size: Optional[tuple[int, int]]
    ) -> Optional[tuple[int, int]]:
        if figure_size is not None:
            return figure_size

        if len(self._benchmark_space) < 3:
            return None

        search_space_values = list(self._benchmark_space.values())[1:]
        total_plots = np.prod([len(v) for v in search_space_values])

        sub_plot_width, sub_plot_height = 5, 4

        n_cols = len(search_space_values[0])
        n_rows = np.ceil(total_plots / n_cols)

        return sub_plot_width * n_cols, sub_plot_height * n_rows

    def _add_significance_to_summary(self, df: pd.DataFrame) -> pd.DataFrame:
        if self._significance_results is None:
            return df

        benchmark_space = list(self.benchmark_space)

        def _make_condition(g: pd.DataFrame, col: str) -> str:
            value = g[col].iloc[0]
            return (
                f"{col} == '{value}'" if isinstance(value, str) else f"{col} == {value}"
            )

        def _function(group: pd.DataFrame):
            group = group.sort_values("mean")
            base_conditions = [_make_condition(group, col) for col in benchmark_space]
            p_values = []
            for i in range(len(group.index) - 1):
                fn1 = group.iloc[i, 0]
                fn2 = group.iloc[i + 1, 0]
                group_conditions = base_conditions + [f"function == '{fn1}'"]
                query_expr = " and ".join(group_conditions)

                # significance results can't be None here
                query_row = self._significance_results.query(query_expr)  # type: ignore
                p_values.append(query_row[fn2].iloc[0])
            p_values.append(np.nan)

            group["p_value"] = p_values

            return group

        groups = df.groupby(benchmark_space)
        result = groups.apply(_function, include_groups=True).reset_index(drop=True)
        columns = (
            ["function", "p_value"] + benchmark_space + ["max", "min", "mean", "std"]
        )
        result = result[columns]

        return result
        # return df

    def _get_max_column_widths(self, df: pd.DataFrame) -> dict[str, int]:
        header_widths = {h: len(h) for h in df.columns}
        data_widths = df.map(lambda x: len(self._stringify_value(x))).max().to_dict()

        return {h: max(hw, data_widths[h]) for h, hw in header_widths.items()}

    def _stringify_value(self, value: Any, width: Optional[int] = None) -> str:
        if isinstance(value, (float, np.floating)):
            raw = "-" if np.isnan(value) else "{:{}}".format(value, self._float_fmt)
            return self._stringify_value(raw, width) if width else raw

        return "{:>{}}".format(str(value), width) if width else str(value)

    def _stringify_header(self, columns: Iterable[str], widths: dict[str, int]) -> str:
        return "  ".join(["{:>{}}".format(h, f"{widths[h]}s") for h in columns])

    def _stringify_group(
        self, group: pd.DataFrame, widths: dict[str, int]
    ) -> list[str]:
        sorted_group = group.sort_values(by=self._sort_by)
        return [self._stringify_row(row, widths) for _, row in sorted_group.iterrows()]

    def _stringify_row(self, row: pd.Series, widths: dict[str, int]) -> str:
        return "  ".join(
            self._stringify_value(v, widths[str(i)]) for i, v in row.items()
        )



"""Utilities for working with pandas."""

# Future Library
from __future__ import annotations

# Standard Library
import logging

# 3rd Party Library
import numpy as np
import pandas as pd

from icontract import require

# 1st Party Library
from rwskit.collections_ import get_first_non_null_value
from rwskit.numpy import get_dtype

log = logging.getLogger(__name__)


def _expand_list(df: pd.DataFrame, column_name: str, string_fill: str) -> pd.DataFrame:
    column = df.pop(column_name)
    min_length = column.map(len).min()
    max_length = column.map(len).max()
    new_names = [f"{column_name}__{i}" for i in range(max_length)]
    new_shape = (len(df.index), max_length)
    dtype = get_dtype(get_first_non_null_value(column))
    is_str = np.issubdtype(dtype, np.str_)
    fill_value = string_fill if is_str else np.nan

    if is_str:
        # max_length is the length of the list, not the max length string
        # in the list.
        max_string_length = max(
            (len(s) if s else 0 for sublist in column for s in sublist)
        )
        dtype = f"U{max(max_string_length, len(string_fill))}"

    if min_length == max_length:
        # We just need to replace 'None`, with the string fill value.
        values = [
            [string_fill if v is None and is_str else v for v in inner_list]
            for inner_list in column.to_list()
        ]
        new_data = np.array(values, dtype=dtype)
    else:
        # Otherwise we need to pad and fill an empty array of the correct
        # dtype.
        # Convert to float so we can fill all missing values with NaN
        if np.issubdtype(dtype, np.number) or dtype == np.bool_:
            dtype = np.float64

        new_data = np.full(new_shape, fill_value=fill_value, dtype=dtype)

        for i, values in enumerate(column):
            # If `None` appears inside a string list, replace it with the
            # fill value. Note, if we don't replace it here, then it
            # will be inserted into the array as the string 'None'.
            values = [fill_value if v is None and is_str else v for v in values]
            new_data[i, : len(values)] = values

    df[new_names] = pd.DataFrame(new_data, index=column.index, columns=new_names)

    return df


@require(
    lambda string_fill: string_fill is not None,
    "The 'string_fill' value cannot be None.",
)
def flatten_data_frame(
    df: pd.DataFrame, string_fill: str = "[UNK]", in_place: bool = False
) -> pd.DataFrame:
    """Converts columns containing lists into (new) individual columns in the ``DataFrame``.

    If one or more columns in a DataFrame consist of lists, this method will
    remove the original column and replace it with ``N`` columns, where
    ``N`` is the maximum length of the lists in the original column.

    If the lists are of unequal length, the additional columns will be appended
    to the right. Lists of strings will be padded using the given
    ``string_fill`` value. All others will be padded with ``np.nan``. Note,
    most numpy types will convert ``np.nan`` into an appropriate missing
    value for that type. For example, when used to fill ``np.datetime64``
    objects, the resulting object will be ``np.datetime64('NaT')``.

    If the lists are numeric (including boolean) and they do not have equal
    lengths, the new columns will have ``dtype=np.float64`` regardless of
    the original dtype.

    .. note::
        Nested lists within a column are not supported and will not be
        flattened.

    Parameters
    ----------
    df : pandas.DataFrame
        The input DataFrame to flatten.
    string_fill : any, default = '[UNK]'
        Use this value to pad string lists. All other data types will use
        ``np.nan``
    in_place : bool, default = False
        Whether to modify the DataFrame in place or return a copy.

    Returns
    -------
    df : pandas.DataFrame
        The modified DataFrame

    Examples
    --------

    .. code-block:: python

        >>>input_df = pd.DataFrame({
            "A": [["1"], ["2", "3"]],
            "B": [["4", "5"], ["6", "7", "8"]],
            "C": [[1], [2, 3]],
            "D": [True, False]
        })
        >>>print(input_df)
                A          B       C      D
        0     [1]     [4, 5]     [1]   True
        1  [2, 3]  [6, 7, 8]  [2, 3]  False

        >>>flatten_data_frame(input_df)
          A__0   A__1 B__0 B__1   B__2  C__0  C__1      D
        0    1  [UNK]    4    5  [UNK]   1.0   NaN   True
        1    2      3    6    7      8   2.0   3.0  False

    """

    if not in_place:
        df = df.copy()

    for c in df.columns:
        if isinstance(df[c].iat[0], list):
            df = _expand_list(df, c, string_fill)

    return df



"""Utilities for working with collections."""

# Future Library
from __future__ import annotations

# Standard Library
import logging as log

from types import GeneratorType
from typing import Any, Iterable, Mapping, Optional

# 3rd Party Library
import yaml

log = log.getLogger(__name__)


def is_iterable(obj: Any, consider_string_iterable: bool = False) -> bool:
    """
    Tests if the object is iterable.

    Parameters
    ----------
    obj : any
        An object.
    consider_string_iterable : bool, default = False
        Whether to consider strings iterable or not.

    Returns
    -------
    bool
        ``True`` if ``obj`` is iterable.
    """
    if isinstance(obj, str):
        return consider_string_iterable

    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True


def is_generator(obj: Any) -> bool:
    """
    Checks if the object is a generator.

    Parameters
    ----------
    obj : any
        An object.

    Returns
    -------
    bool
        ``True`` if the object is a generator.

    """
    return isinstance(obj, GeneratorType)


def get_first_non_null_value(collection: Iterable) -> Optional[Any]:
    """Recursively try to get the first non-null value in a collection.

    This method will recursively traverse the collection until it finds
    a non-iterable value that is not ``None``.

    Parameters
    ----------
    collection : Iterable
        The collection to retrieve the value from.

    Returns
    -------
    Any, optional
        The first non-null value in the series, if one exists, otherwise return
        ``None``.

    """
    for value in collection:
        if is_iterable(value) and not isinstance(value, str):
            value = get_first_non_null_value(value)

        if value is not None:
            return value

    return None


def recursive_sort(obj: Any) -> Any:
    """
    Attempts to sort an object recursively according to the following rules:

    * If the object is a dictionary, it will be sorted by its keys and its
      values will be sorted recursively.
    * If the object is a list or tuple, it will be sorted by its values.
      Typically, a value is a primitive, but lists, tuples, and dictionaries
      are also valid. In these cases, the collections are compared after
      sorting.
    * Any other value is returned unchanged.

    .. note::
        Collections of dictionaries (e.g., lists, tuples, sets, other
        dictionaries) cannot be directly compared because ``__lt__`` is not
        defined. When a collection of dictionaries is encountered, they are
        compared using (sorted) yaml strings. Yaml is used because it supports
        sorting by keys, similar to json, but supports more data types.

    Parameters
    ----------
    obj : Any
        The input object

    Returns
    -------
    Any
        The returned object sorted.

    Raises
    ------
    TypeError
        If you have a collection of dictionaries that cannot be converted to
        yaml strings.
    """

    def _key_fn(e: Any) -> Any:
        if isinstance(e, dict):
            return yaml.dump(
                e,
                sort_keys=True,
                default_flow_style=True,
                explicit_start=False,
                explicit_end=False,
                width=float("inf"),
            )
        elif is_iterable(e):
            return sorted(e, key=lambda x: _key_fn(x))
        else:
            return e

    if isinstance(obj, dict):
        return {k: recursive_sort(obj[k]) for k in sorted(obj)}
    elif isinstance(obj, list):
        values = sorted(obj, key=lambda e: _key_fn(e))
        return [recursive_sort(e) for e in values]
    elif isinstance(obj, tuple):
        values = sorted(obj, key=_key_fn)
        return tuple([recursive_sort(e) for e in values])
    elif is_iterable(obj) or is_generator(obj):
        return list(obj)
    else:
        return obj


def remove_none_from_dict(d: dict[Any, Any]) -> dict[Any, Any]:
    """Recursively remove ``None`` values from a (nested) dictionary."""

    def _do_removal(obj: Any) -> Any:
        if isinstance(obj, Mapping):
            return {k: _do_removal(v) for k, v in obj.items() if v is not None}
        elif isinstance(obj, list):
            return [remove_none_from_dict(e) for e in obj]
        elif isinstance(obj, set):
            return {remove_none_from_dict(e) for e in obj}  # type: ignore
        else:
            return obj

    return _do_removal(d)



# Future Library
from __future__ import annotations

# Standard Library
import abc
import dataclasses
import datetime
import json
import logging
import os
import re

from contextlib import contextmanager
from io import IOBase, StringIO
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Self,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

# 3rd Party Library
import yaml
import yaml_include

from dateutil.parser import parse as parse_date
from icontract import ensure, require
from pydantic.dataclasses import dataclass
from yamlable import YamlAble

# 1st Party Library
from rwskit.dataclasses_ import DataclassRegistry
from rwskit.types_ import get_qualified_name

log = logging.getLogger(__name__)


T = TypeVar("T", bound="YamlConfig")
E = TypeVar("E", bound="EnvironmentConfig")
I = TypeVar("I")

TypeParser = Callable[[str], I]


# Enable include directives inside yaml files
@contextmanager
def _enable_yaml_include(path: Path):
    tags = ("!inc", "!include")
    loaders = (yaml.Loader, yaml.BaseLoader, yaml.SafeLoader, yaml.FullLoader)
    base_dir = path.parent

    for tag in tags:
        for loader in loaders:
            yaml.add_constructor(
                tag, yaml_include.Constructor(base_dir=base_dir), loader
            )

    yield

    for tag in tags:
        for loader in loaders:
            del loader.yaml_constructors[tag]


# According to the python documentation the following (commented) code should
# allow any class that uses the 'YamlConfigMeta' to automatically be viewed
# as a dataclass from the perspective of the type checker. Unless I am doing
# something wrong (very possible), it does not work for me and PyCharm
# does not detect that subclasses should be considered as dataclasses. The
# code functions correctly, but PyCharm will issue warnings about unresolved
# attributes for all subclasses and cannot provide autocompletion.
#
# However, if you annotate each individual class with @dataclass_transform
# PyCharm will recognize the class as a dataclass. This is obviously not ideal
# but better than nothing.

# @dataclass_transform(kw_only_default=True, frozen_default=True)
# class DataclassMeta(type):
#     def __new__(mcs, name, bases, namespace, **kwargs):
#         new_cls = super().__new__(name, bases, namespace, **kwargs)
#         return pydantic_dataclass(new_cls, frozen=True, kw_only=True)
#
#
# class YamlConfigMeta(DataclassMeta, type(YamlAble)):
#     pass


class YamlConfig(YamlAble):
    """A base class for serializable configuration objects.

    Classes that inherit from this class can easily be serialized to and from
    YAML files. Given a YAML file, the class can be reconstructed as long
    as the YAML attributes can be uniquely mapped to a subclass of
    :class:`YamlConfig`.

    Additionally, the configuration can be split across multiple files for
    better modularity using the ``!include`` directive.

    Examples
    --------
    >>> class ChildConfig(YamlConfig):
    ...     id: int
    ...     name: str
    ...     timestamp: datetime.datetime
    >>> class ParentConfig(YamlConfig):
    ...     parent_attr: str = "parent_attr_value"
    ...     child_attr:
    >>> expected_config = ParentConfig(
    ...     id=1,
    ...     child_config=ChildConfig(
    ...         id=2,
    ...         name="child_config",
    ...         timestamp=datetime.datetime.now()
    ... )
    >>> plain_yaml = '''
    ... child_config:
    ...     id: 2
    ...     name: child_config
    ...     timestamp: 2024-11-19 13:55:34.064388
    ...  id: 1
    ... '''
    >>> from_plain_yaml = YamlConfig.loads_yaml(plain_yaml)
    >>> assert from_plain_yaml == expected_config

    The ``!yamlable`` tag can be used  to explicitly tell the YAML parser
    which class to construct. The syntax is ``!yamlable/<fully_qualified_class_name>``.

    >>> tagged_yaml = '''
    ... !yamlable/my_package.my_module.ParentConfig
    ... child_config: !yamlable/my_package.my_module.ChildConfig
    ...     id: 2
    ...     name: child_config
    ...     timestamp: 2024-11-19 13:55:34.064388
    ... id: 1
    ... '''
    >>> assert YamlConfig.loads_yaml(tagged_yaml) == expected_config

    You can use the ``!include`` directive to include other YAML files.
    For example, assume you have the following two YAML files:

    .. code-block:: yaml

        # child_config.yaml

        id: 2
        name: "child_config"
        timestamp: 2024-11-19 13:55:34.064388

    .. code-block:: yaml

        # parent_config.yaml

        id: 1
        child_config: !include child_config.yaml

    You can load the parent config using ``YamlConfig.load_yaml`` as follows:

    >>> YamlConfig.load_yaml("parent_config.yaml")
    """

    # Maintain a registry of all classes that inherit from this base class.
    # Although, I don't plan to use this in a multithreaded context, it is
    # pretty easy to make it thread safe with a lock.
    __registry: ClassVar[DataclassRegistry] = DataclassRegistry()

    default_type_parsers = {
        int: int,
        float: float,
        bool: lambda x: x.lower() in ("true", "1", "t", "y", "yes"),
        str: lambda x: x,
        datetime.datetime: parse_date,
        datetime.date: lambda x: parse_date(x).date(),
    }

    @require(
        lambda self: dataclasses.is_dataclass(self), "The class must be a dataclass."
    )
    def __init__(self):
        pass

    def __init_subclass__(cls: Type[Self], **kwargs: Any):
        """Initialize subclasses to make them suitable configuration objects.

        * Automatically assign the ``__yaml_tag_suffix__`` using the fully
          qualified class name.
        * Convert the class to a ``pydantic.dataclasses.dataclass`` that has
          ``frozen=True`` and ``kw_only=True``.

        Parameters
        ----------
        kwargs : Any
        """
        super().__init_subclass__(**kwargs)

        cls.__yaml_tag_suffix__ = get_qualified_name(cls)

        # Add the subclass to the registry so we can dynamically load the class
        # without having to import it first.
        YamlConfig.__registry.register(cls)  # noqa

    @classmethod
    def get_registered_classes(cls) -> set[Type[YamlConfig]]:
        """Get the set of classes currently registered as configuration objects.

        Returns
        -------
        set[Type[YamlConfig]]
            The set of registered yaml config classes.

        """
        return set(cls.__registry)

    def dumps_plain_yaml(self: Self) -> str:
        """
        Represent the class as plain YAML without any tags.

        .. note::
            It may not be possible to reconstruct the python object from this
            string.

        Returns
        -------
        str
            The object as plain YAML without any tags.
        """
        d = self._transform_sets_to_lists(self._dataclass_to_dict(self))

        return yaml.safe_dump(d)

    @classmethod
    def _dataclass_to_dict(cls, obj: Any) -> Any:
        if dataclasses.is_dataclass(obj) and not isinstance(obj, type):
            return {
                k: cls._dataclass_to_dict(v) for k, v in dataclasses.asdict(obj).items()
            }
        elif isinstance(obj, str):
            return obj
        elif isinstance(obj, Mapping):
            return {k: cls._dataclass_to_dict(v) for k, v in obj.items()}
        elif isinstance(obj, Iterable):
            return [cls._dataclass_to_dict(v) for v in obj]
        else:
            return obj

    @classmethod
    def _transform_sets_to_lists(cls, obj: Any) -> Any:
        if isinstance(obj, set):
            return [cls._transform_sets_to_lists(e) for e in obj]
        elif isinstance(obj, Mapping):
            return {k: cls._transform_sets_to_lists(v) for k, v in obj.items()}
        elif isinstance(obj, str):
            return obj
        elif isinstance(obj, Iterable):
            return [cls._transform_sets_to_lists(v) for v in obj]
        else:
            return obj

    @classmethod
    @require(
        lambda file_path_or_stream: isinstance(
            file_path_or_stream, (str, Path, IOBase, StringIO)
        ),
        "'file_path_or_stream' must be a string, pathlib.Path, IOBase, or StringIO object.",
    )
    def load_yaml(
        cls: Type[T],
        file_path_or_stream: str | Path | IOBase | StringIO,
        safe: bool = True,
    ) -> T:
        raw_config = cls._load_raw_yaml(file_path_or_stream, safe)

        # If the deserialized object is an instance of the loader class then
        # we can simply return the object (i.e., it contained all the tags
        # necessary to reconstruct the config).
        if isinstance(raw_config, cls):
            return raw_config

        # Since a config file represents a data class the top level
        # object of a plain yaml config must be a dictionary (i.e., the
        # attributes of the dataclass).
        if not isinstance(raw_config, Mapping):
            raise TypeError("Invalid YAML config file.")

        return YamlConfig.__registry.construct_registered_dataclass(raw_config)

    @classmethod
    def _load_raw_yaml(
        cls: Type[T],
        file_path_or_stream: str | Path | IOBase | StringIO,
        safe: bool = True,
    ) -> Any:
        # Deserialize the yaml into an object. If the yaml contains known
        # tags it will return the appropriate python classes, otherwise
        # it will return primitive python types (e.g., ints, floats, lists,
        # dicts, etc.)

        if not safe:
            raise NotImplementedError("Only safe loading is supported.")

        yaml_loader = yaml.safe_load

        if isinstance(file_path_or_stream, (str, Path)):
            with open(file_path_or_stream, "rt") as fh:
                # This allows using paths relative to the input config file
                # rather than paths relative to the current working directory.
                with _enable_yaml_include(Path(file_path_or_stream)):
                    return yaml_loader(fh)
        else:
            with file_path_or_stream as fh:
                return yaml_loader(fh)


@dataclass(frozen=True, kw_only=True, config={"arbitrary_types_allowed": True})
class EnvironmentConfig(abc.ABC):
    """A mixin class for configuration objects to help constructing them from environment variables.

    This mixin adds a method :meth:`from_environment` that will try to parse
    environment variables into the correct type for the dataclass. All you
    have to do is implement the :meth:`environment_mapping` method to provide
    a mapping between environment variable names and the dataclass field names.
    """

    _default_type_parsers: ClassVar[dict[Any, Any]] = {
        bool: lambda x: x.lower() in ("true", "1", "t", "y", "yes"),
        int: int,
        float: float,
        str: lambda x: x,
        Literal: lambda x: x,
        datetime.datetime: parse_date,
        datetime.date: lambda x: parse_date(x).date(),
        re.Pattern: lambda x: re.compile(x),
    }
    _dataclass_fields = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def get_default_type_parsers(cls) -> dict[Any, TypeParser]:
        """Get the default type parsers for this class."""

        return cls._default_type_parsers.copy()

    @classmethod
    @abc.abstractmethod
    @ensure(
        lambda cls, result: all(cls._is_valid_field(v) for v in result.values()),
        "All values in the mapping must be a valid field name.",
    )
    def environment_mapping(cls: Type[E]) -> Mapping[str, str]:
        """
        Returns a mapping from environment variable names to their
        corresponding field name.

        Returns
        -------
        Mapping[str, str]

        Raises
        ------
        ViolationError
            If any of the returned values are not a field name of this
            class. Note, not all fields need to be included in the mapping, but
            any entry that is included must correspond to a field name.
        """
        pass

    @classmethod
    def _is_valid_field(cls, field_name: str) -> bool:
        # A valid field is any field that can be used in the constructor.
        field_names = {f.name for f in dataclasses.fields(cls) if f.init}  # type: ignore

        return field_name in field_names

    @classmethod
    def from_environment(
        cls, type_parsers: Optional[dict[Any, TypeParser]] = None, **kwargs
    ) -> Self:
        """
        Create an instance from environment variables.

        Environment variables can represent python primitive values,
        date objects, datetime objects. They can also be lists, sets, tuples,
        or dicts of these types. Collections are space separated string.
        Multiword strings should be enclosed in double quotes. Dictionary
        key value pairs are delimited by '='. Spaces are not allowed in
        keys or values.

        Parameters
        ----------
        type_parsers: dict[Any, TypeParser]
            A mapping from python types to functions that parse a string
            into that type. These will be combined and override the default
            parsing rules described above.

        kwargs
            Additional keyword arguments to override the values from the
            environment or composite fields (e.g., other dataclasses),
            which are constructed externally.

        Returns
        -------

        """
        type_parsers = cls.get_default_type_parsers() | (type_parsers or {})
        environment_kwargs = cls.get_environment_kwargs(type_parsers)
        init_kwargs = environment_kwargs | kwargs

        return cls(**init_kwargs)

    @classmethod
    def get_environment_kwargs(
        cls, type_parsers: Optional[dict[Any, TypeParser]] = None
    ) -> dict[str, Any]:
        """
        Get the keyword arguments for this dataclass that are available
        from the environment.

        Parameters
        ----------
        type_parsers : dict[Type, TypeParser]
            The type parsers to use when parsing environment variables.

        Returns
        -------
        dict[str, Any], optional
            The keyword arguments as a dictionary.

        """
        type_parsers = cls.get_default_type_parsers() | (type_parsers or {})

        # Map the fields to the environment variable (string) values, for all
        # non-empty values.
        value_map = {
            fn: ev
            for vn, fn in cls.environment_mapping().items()
            if (ev := os.environ.get(vn)) is not None  # noqa
        }

        # For each field, attempt to parse the value.
        resolved_types = get_type_hints(cls)
        return {
            fn: cls.parse_value_from_string(
                ev, cls._get_actual_type(resolved_types[fn]), type_parsers
            )
            for fn, ev in value_map.items()
        }

    # A utility method to make sure we get the actual field type in case
    # it is marked optional.
    @classmethod
    def _get_actual_type(cls, t: Type[I]) -> Type[I]:
        if get_origin(t) is Union:
            args = get_args(t)
            if len(args) == 2 and type(None) in args:
                args = [a for a in args if a is not type(None)]
                return args[0]
            else:
                raise NotImplementedError(
                    "Unions of more than one type are not supported"
                )
        return t

    @classmethod
    def get_dataclass_field_map(cls: Type[Self]) -> dict[str, dataclasses.Field]:
        """Get a mapping from field names to their dataclass fields."""

        if cls._dataclass_fields is None:
            cls._dataclass_fields = {f.name: f for f in dataclasses.fields(cls)}  # type: ignore

        return cls._dataclass_fields

    @classmethod
    def get_field_from_name(cls, field_name: str) -> dataclasses.Field:
        """Get the dataclass field from its name."""
        return cls.get_dataclass_field_map()[field_name]

    @classmethod
    def parse_value_from_string(
        cls,
        input_string: str,
        target_type: Type[I],
        type_parsers: Optional[dict[Type[I], TypeParser]] = None,
    ):
        """
        Try to parse an input string into the given python ``target_type``.

        The input string can be a single value that can be parsed by
        the default parsing rules or any of the rules implemented by the
        given ``type_parsers``. By default, the parseable values are: str, int,
        float, bool, datetime.datetime, datetime.date, re.Pattern, and any
        class that can be constructed from a single string argument (e.g.,
        ``pathlib.Path``.

        To handle more complex ``target_types``, the input string can also
        be a JSON string. The string will be parsed as JSON and the resulting
        object will be traversed to try to convert the leaf values using the
        all the available parsing rules.

        Parameters
        ----------
        input_string : str
            The string to parse.
        target_type : Type
            The python type to parse the value into.
        type_parsers : dict[Type, TypeParser], optional
            An optional dictionary that maps a type to a function that
            parses a string to that type. These rules will augment the default
            parsing rules given by :meth:`get_default_type_parsers` and will
            take precedent over the default rules if there is a conflict.

        Returns
        -------
        An instance of the target type parsed from the environment value.
        """
        type_parsers = cls.get_default_type_parsers() | (type_parsers or {})

        def parse_base_type(value: str, hint: Type[Any]):
            """Parse a base type (no origin) from a string."""

            def constructor_parser(x: str) -> Any:
                """Try parsing ``x`` assuming ``hint`` is callable"""
                # If hint is not callable, we'll catch the exception later.
                return hint(x)  # type: ignore

            hint = get_origin(hint) or hint

            try:
                return type_parsers.get(hint, constructor_parser)(value)
            except Exception as e:
                raise ValueError(
                    f"Unable to parse '{value}' of type '{type(value)}' as {hint}. "
                    f"Raised from {type(e)}: {str(e)}"
                ) from e

        def parse_value(value: Any, hint: Type[Any]):
            """Recursively try to parse an arbitrary value to the given type."""
            origin = get_origin(hint)

            if origin is Literal:
                return parse_value(value, str)
            if origin is list:
                inner_type = get_args(hint)[0]
                return [parse_value(e, inner_type) for e in value]
            elif origin is set:
                inner_type = get_args(hint)[0]
                return set(parse_value(e, inner_type) for e in value)
            elif origin is tuple:
                inner_types = get_args(hint)
                return tuple(parse_value(e, t) for e, t in zip(value, inner_types))
            elif origin is dict:
                key_type, value_type = get_args(hint)
                return {
                    parse_value(k, key_type): parse_value(v, value_type)
                    for k, v in value.items()
                }
            elif origin is Union:
                inner_types = [t for t in get_args(hint) if t is not type(None)]
                for inner_type in inner_types:
                    try:
                        return parse_value(value, inner_type)
                    except Exception:
                        pass
                raise ValueError(f"Unable to parse '{value}' as an of {inner_types}.")
            elif origin is None:
                value_type = type(value)
                if value_type in (int, float, bool):
                    # Note, I shouldn't have to also check that the target_type
                    # matches the value_type, because pydantic will do that
                    # for us when it constructs the object.
                    return value
                elif value_type is str:
                    return parse_base_type(str(value), hint)
                else:
                    raise ValueError(
                        f"Unable to parse '{value}' of type '{value_type}' as {hint}."
                    )

            raise ValueError(f"Unsupported type: {hint}")

        try:
            return parse_value(json.loads(input_string), target_type)
        except json.JSONDecodeError:
            return parse_base_type(input_string, target_type)



# Future Library
from __future__ import annotations

# Standard Library
import logging

from typing import List, Union


class LogLevel(int):
    """A class for converting logging level string names to their integer counterparts."""

    def __new__(cls, level: Union[str, int]):
        if isinstance(level, int):
            if level in logging._levelToName:
                return super().__new__(cls, level)
            else:
                raise TypeError(f"Invalid logging level: {level}")

        # Otherwise assume we've got a string.
        level = level.upper()

        try:
            return super().__new__(cls, logging._nameToLevel[level])
        except KeyError:
            raise TypeError(f"invalid logging level: {level}")

    def __repr__(self):
        return logging._levelToName[self]

    def __str__(self):
        return self.__repr__()

    @classmethod
    def valid_levels(cls) -> List[LogLevel]:
        """
        Returns the list of valid log levels.

        Returns
        -------
        list[LogLevel]
            The list of valid log levels.
        """
        return [LogLevel(name) for name in logging.getLevelNamesMapping()]



"""Utilities for numeric operations."""

# Future Library
from __future__ import annotations

# Standard Library
import ctypes

from typing import Optional

int_8_max = (1 << 8) - 1
int_16_max = (1 << 16) - 1
int_32_max = (1 << 32) - 1
int_64_max = (1 << 64) - 1


def int_(s: Optional[str], falsy_as_none: bool = True) -> Optional[int]:
    """
    Convert a string `s` to an integer. If `s` is `None` return `None`
    instead of raising an exception.

    Parameters
    ----------
    s : str, optional
        The string to convert.
    falsy_as_none : bool, default=True
        If `True` all falsy values of `s` (other than ``0``) will return
        `None`, otherwise only `None` values will be passed through and
        an exception will be raised.

    Returns
    -------
    int, optional
        `None` if `s` is `None`, otherwise an integer if `s` is not `None` and
        is parseable.

    Raises
    ------
    ValueError
        If the string cannot be parsed.


    """
    # Note `int` strips whitespace
    if falsy_as_none:
        return int(s) if s else None

    return int(s) if s is not None else None


def float_(s: Optional[str], falsy_as_none: bool = True) -> Optional[float]:
    """
    Convert a string `s` to a float. If `s` is `None` return `None`
    instead of raising an exception.

    Parameters
    ----------
    s : str, optional
        The string to convert.
    falsy_as_none : bool, default=True
        If `True` all falsy values of `s` will return `None` (other than ``0``),
        otherwise only `None` values will be passed through and an exception
        will be raised..

    Returns
    -------
    float, optional
        `None` if `s` is `None`, otherwise a float if `s` is not `None` and
        is parseable.

    Raises
    ------
    ValueError
        If the string cannot be parsed.


    """
    if falsy_as_none:
        return float(s) if s else None

    return float(s) if s is not None else None


def to_signed(value: int) -> int:
    """
    If the value is positive, this will return a signed version of the number
    that fits in the same size integer type. For example,  ``129`` is a valid
    8-bit signed integer, but a 16-bit unsigned integer. This method will
    convert the number to the corresponding signed integer that still fits
    in 8-bits. In this case ``127``.

    Parameters
    ----------
    value : int
        The value to convert.

    Returns
    -------
    int
        The signed version.

    Raises
    ------
    ValueError
        If the ``value`` is more than 64-bits.
    """
    if value > int_64_max:
        raise ValueError(f"Only 64-bit values or smaller are supported. Got: {value}")

    if value <= 0:
        return value

    if value <= int_8_max:
        return ctypes.c_int8(value).value
    if value <= int_16_max:
        return ctypes.c_int16(value).value
    if value <= int_32_max:
        return ctypes.c_int32(value).value
    if value <= int_64_max:
        return ctypes.c_int64(value).value

    raise ValueError(f"Unexpected value: {value}")



"""String utilities."""

# Future Library
from __future__ import annotations


def camel_to_snake_case(s: str) -> str:
    """Convert the string from CamelCase to snake_case.

    This method uses a simple regular expression to convert a string
    in CamelCase to one in snake_case. Strings with 2 or fewer
    are simply lowercased and returned.

    Parameters
    ----------
    s : str
        The string to convert.

    Returns
    -------
    str
        The string in snake case.
    """
    # There are a number of approaches for converting CamelCase to snake_case
    # enumerated at
    # https://www.geeksforgeeks.org/python-program-to-convert-camel-case-string-to-snake-case/
    # However, the only one that passes all my unit tests is method #4 (regex).
    # The "naive" approach below also passes all the unit tests, is also
    # faster than any of the other listed methods (while #4 is the slowest).
    # See this post for some additional regex approaches:
    # https://stackoverflow.com/a/1176023
    if not s:
        return s

    if len(s) <= 2:
        return s.lower()

    # What will be the previous and next character
    p, c = s[0], s[1]

    # The array to store the result in.
    snake = [p]

    # The iterated variable represents the next character
    for n in s[2:]:
        # If the current character is uppercase and the next character
        # is not, then we'll start a new "word"
        if p.isalpha() and c.isupper() and not n.isupper():
            snake.append("_")

        snake.append(c)

        # Update the previous and next characters
        p, c = c, n

    snake.append(c)

    return "".join(snake).lower()
